@@ -74,12 +74,11 @@ class IsConnected(object):
             return False
         return True
 
-    def test_unconnected(self, dataset_settings, local_dataset=''):
+    def test_unconnected(self, replicate_param, local_dataset=''):
         """
         Check that endpoint is unconnected
         """
         self.local_dataset = local_dataset
-        replicate_param = dataset_settings['replicate']
         if (replicate_param and replicate_param['endpoint_host']):
             host = replicate_param['endpoint_host']
             port = replicate_param['endpoint_port']
@@ -142,7 +141,7 @@ class Manager(object):
                     if take_snapshot is True or replicate is True or clean is True:
                         if dataset_settings['time'].is_trigger():
                             # Check endpoint for trigger is connected
-                            if test_reachable and is_connected.test_unconnected(dataset_settings):
+                            if test_reachable and is_connected.test_unconnected(dataset_settings['replicate']):
                                 continue
                             # Trigger file testing and creation
                             trigger_filename = '{0}/{1}'.format(dataset_settings['mountpoint'], TRIGGER_FILENAME)
@@ -178,6 +177,18 @@ class Manager(object):
             result = PROC_CHANGED
         return result
 
+    @staticmethod
+    def new_hold(dataset, snap_name, endpoint='', log_command=False):
+        result = PROC_EXECUTED
+        holds = ZFS.holds(dataset, endpoint=endpoint, log_command=log_command)
+        ZFS.hold(dataset, snap_name, endpoint=endpoint, log_command=log_command, may_exist=True)
+        if snap_name in holds:
+            holds.remove(snap_name)
+        for hold in holds:
+            ZFS.release(dataset, hold, endpoint=endpoint, log_command=log_command)
+        result = PROC_CHANGED
+        return result
+
     @staticmethod
     def replicate(src_dataset, src_snapshots, dst_dataset, dst_snapshots, replicate_settings):
         result = PROC_EXECUTED
@@ -186,19 +197,55 @@ class Manager(object):
         src_endpoint = '' if push else replicate_settings['endpoint']
         dst_endpoint = replicate_settings['endpoint'] if push else ''
         src_host = gethostname().split('.')[0] if push else replicate_settings['endpoint_host']
+        dst_host = gethostname().split('.')[0] if not push else replicate_settings['endpoint_host']
         local_dataset = src_dataset if push else dst_dataset
         full_clone = replicate_settings['full_clone']
+        receive_save = replicate_settings['receive_save']
+        receive_no_mountpoint = replicate_settings['receive_no_mountpoint']
+        receive_umount = replicate_settings['receive_umount']
         send_compression = replicate_settings['send_compression']
         send_properties = replicate_settings['send_properties']
+        send_raw = replicate_settings['send_raw']
         all_snapshots = replicate_settings['all_snapshots']
         buffer_size = replicate_settings['buffer_size']
         compression = replicate_settings['compression']
         log_command = replicate_settings['log_commands']
         extra_args = {'full_clone': full_clone, 'all_snapshots': all_snapshots,
+                'receive_no_mountpoint': receive_no_mountpoint, 'receive_umount': receive_umount,
+                'receive_save': receive_save,
                 'send_compression': send_compression, 'send_properties': send_properties,
-                'buffer_size': buffer_size, 'compression': compression,
+                'buffer_size': buffer_size, 'compression': compression, 'send_raw': send_raw,
                 'log_command': log_command }
-        log_info('[{0}] - Replicating [{1}]:{2}'.format(local_dataset, src_host, src_dataset))
+        # Get any receive_resume_tokens
+        receive_resume_token = ''
+        if push:
+            receive_resume_token = ZFS.get_receive_resume_token(dst_dataset, endpoint=dst_endpoint, log_command=log_command)
+        else:
+            receive_resume_token = ZFS.get_receive_resume_token(src_dataset, endpoint=src_endpoint, log_command=log_command)
+        if receive_resume_token:
+            log_info('[{0}] - Resuming replicating [{1}]:{2} to [{3}]:{4}'.format(local_dataset, src_host, src_dataset, dst_host, dst_dataset))
+            size = ZFS.get_size(src_dataset, None, None, src_endpoint, receive_resume_token, **extra_args)
+            log_info('[{0}] -   {1}@??? > {1}@??? ({2})'.format(local_dataset, src_dataset, size))
+            ZFS.replicate(src_dataset, None, None, dst_dataset, replicate_settings['endpoint'],
+                    receive_resume_token, direction=replicate_dirN, **extra_args)
+            # Recalculate dst data sets
+            if push:
+                dst_endpoint = replicate_settings['endpoint']
+            else:
+                dst_endpoint = ''
+            new_dst_snapshots = ZFS.get_snapshots2(dst_dataset, dst_endpoint, log_command=log_command,
+                    all_snapshots=all_snapshots)
+            snapshot = list(new_dst_snapshots)[-1]
+            snap_name = new_dst_snapshots[snapshot]['name']
+            Manager.new_hold(src_dataset, snap_name, endpoint=src_endpoint, log_command=log_command)
+            Manager.new_hold(dst_dataset, snap_name, endpoint=dst_endpoint, log_command=log_command)
+            dst_snapshots.clear()
+            dst_snapshots.update(new_dst_snapshots)
+            log_info('[{0}] - Resumed replicatiion [{1}]:{2} to [{3}]:{4} complete'.format(local_dataset, src_host, src_dataset, dst_host, dst_dataset))
+            result= PROC_CHANGED
+            return result
+
+        log_info('[{0}] - Replicating [{1}]:{2} to [{3}]:{4}'.format(local_dataset, src_host, src_dataset, dst_host, dst_dataset))
         last_common_snapshot = None
         index_last_common_snapshot = None
         # Search for the last src snapshot that is available in dst
@@ -220,10 +267,8 @@ class Manager(object):
                 log_info('[{0}] -   {1}@{2} > {1}@{3} ({4})'.format(local_dataset, src_dataset, prevsnap_name, snap_name, size))
                 ZFS.replicate(src_dataset, prevsnap_name, snap_name, dst_dataset, replicate_settings['endpoint'],
                         direction=replicate_dirN, **extra_args)
-                ZFS.hold(src_dataset, snap_name, endpoint=src_endpoint, log_command=log_command)
-                ZFS.hold(dst_dataset, snap_name, endpoint=dst_endpoint, log_command=log_command)
-                ZFS.release(src_dataset, prevsnap_name, endpoint=src_endpoint, log_command=log_command)
-                ZFS.release(dst_dataset, prevsnap_name, endpoint=dst_endpoint, log_command=log_command)
+                Manager.new_hold(src_dataset, snap_name, endpoint=src_endpoint, log_command=log_command)
+                Manager.new_hold(dst_dataset, snap_name, endpoint=dst_endpoint, log_command=log_command)
                 for snapshot in snaps_to_send:
                     dst_snapshots.update({snapshot:src_snapshots[snapshot]})
                 result = PROC_CHANGED
@@ -236,10 +281,8 @@ class Manager(object):
                     log_info('[{0}] -   {1}@{2} > {1}@{3} ({4})'.format(local_dataset, src_dataset, prevsnap_name, snap_name, size))
                     ZFS.replicate(src_dataset, prevsnap_name, snap_name, dst_dataset, replicate_settings['endpoint'],
                             direction=replicate_dirN, **extra_args)
-                    ZFS.hold(src_dataset, snap_name, endpoint=src_endpoint, log_command=log_command)
-                    ZFS.hold(dst_dataset, snap_name, endpoint=dst_endpoint, log_command=log_command)
-                    ZFS.release(src_dataset, prevsnap_name, endpoint=src_endpoint, log_command=log_command)
-                    ZFS.release(dst_dataset, prevsnap_name, endpoint=dst_endpoint, log_command=log_command)
+                    Manager.new_hold(src_dataset, snap_name, endpoint=src_endpoint, log_command=log_command)
+                    Manager.new_hold(dst_dataset, snap_name, endpoint=dst_endpoint, log_command=log_command)
                     previous_snapshot = snapshot
                     dst_snapshots.update({snapshot:src_snapshots[snapshot]})
                     result = PROC_CHANGED
@@ -251,7 +294,7 @@ class Manager(object):
             log_info('  {0}@         > {0}@{1} ({2})'.format(src_dataset, snap_name, size))
             ZFS.replicate(src_dataset, None, snap_name, dst_dataset, replicate_settings['endpoint'],
                     direction=replicate_dirN, **extra_args)
-            ZFS.hold(src_dataset, snap_name, endpoint=src_endpoint, log_command=log_command)
+            Manager.new_hold(src_dataset, snap_name, endpoint=src_endpoint, log_command=log_command)
             ZFS.hold(dst_dataset, snap_name, endpoint=dst_endpoint, log_command=log_command)
             if full_clone:
                 for snapshot in src_snapshots:
@@ -259,7 +302,7 @@ class Manager(object):
             else:
                 dst_snapshots.update({snapshot:src_snapshots[snapshot]})
             result = PROC_CHANGED
-        log_info('[{0}] - Replicating [{1}]:{2} complete'.format(local_dataset, src_host, src_dataset))
+        log_info('[{0}] - Replicating [{1}]:{2} to [{3}]:{4} complete'.format(local_dataset, src_host, src_dataset, dst_host, dst_dataset))
         return result
 
     @staticmethod
@@ -284,14 +327,17 @@ class Manager(object):
                 dataset_settings = ds_settings[dataset]
                 take_snapshot = dataset_settings['snapshot'] is True
                 replicate = dataset_settings['replicate'] is not None
+                replicate2 = dataset_settings['replicate2'] is not None
                 clean = bool(dataset_settings['schema'])
 
                 # Decide whether we need to handle this dataset
-                if not take_snapshot and not replicate and not clean:
+                if not take_snapshot and not replicate and not replicate2 and not clean:
                     continue
- 
+
                 replicate_settings = dataset_settings['replicate']
+                replicate2_settings = dataset_settings['replicate2']
                 full_clone = replicate_settings['full_clone'] if replicate else False
+                full_clone2 = replicate2_settings['full_clone'] if replicate2 else False
                 log_command = dataset_settings['log_commands']
                 local_snapshots = snapshots.get(dataset, OrderedDict())
                 # Manage what snapshots we operate on - everything or zsnapd only
@@ -324,31 +370,48 @@ class Manager(object):
                             Helper.run_command(dataset_settings['postexec'], '/', log_command=log_command)
 
                     # Replicating, if required
-                    # If network replicating, check connectivity here
-                    test_unconnected = is_connected.test_unconnected(dataset_settings, local_dataset=dataset)
-                    if test_unconnected:
-                        log_info("[{0}] - Skipping as '{1}:{2}' unreachable"
-                                .format(dataset, replicate_settings['endpoint_host'], replicate_settings['endpoint_port']))
-                        continue
-
+                    result = PROC_FAILURE
+                    result2 = PROC_FAILURE
                     if (replicate is True):
+                        # If network replicating, check connectivity here
+                        test_unconnected = is_connected.test_unconnected(replicate_settings, local_dataset=dataset)
+                        if test_unconnected:
+                            log_info("[{0}] - Skipping as '{1}:{2}' unreachable"
+                                    .format(dataset, replicate_settings['endpoint_host'], replicate_settings['endpoint_port']))
+                            continue
+
                         remote_dataset = replicate_settings['target']
-                        remote_snapshots = ZFS.get_snapshots(remote_dataset, replicate_settings['endpoint'], log_command=log_command,
+                        remote_snapshots = ZFS.get_snapshots2(remote_dataset, replicate_settings['endpoint'], log_command=log_command,
                                 all_snapshots=dataset_settings['all_snapshots'])
-                        remote_snapshots = remote_snapshots.get(remote_dataset, OrderedDict())
                         result = Manager.replicate(dataset, local_snapshots, remote_dataset, remote_snapshots, replicate_settings)
                         # Clean snapshots remotely if one has been taken - only kept snapshots will allow aging
                         if (dataset_settings['remote_schema']):
                             Cleaner.clean(remote_dataset, remote_snapshots, dataset_settings['remote_schema'], log_command=log_command,
                                     all_snapshots=dataset_settings['remote_clean_all'])
-                        # Post execution command
-                        if (result and dataset_settings['replicate_postexec'] is not None):
-                            Helper.run_command(dataset_settings['replicate_postexec'], '/', log_command=log_command)
+                    if (replicate2 is True):
+                        # If network replicating, check connectivity here
+                        test_unconnected = is_connected.test_unconnected(replicate2_settings, local_dataset=dataset)
+                        if test_unconnected:
+                            log_info("[{0}] - Skipping as '{1}:{2}' unreachable"
+                                    .format(dataset, replicate2_settings['endpoint_host'], replicate2_settings['endpoint_port']))
+                            continue
+
+                        remote_dataset = replicate2_settings['target']
+                        remote_snapshots = ZFS.get_snapshots2(remote_dataset, replicate2_settings['endpoint'], log_command=log_command,
+                                all_snapshots=dataset_settings['all_snapshots'])
+                        result2 = Manager.replicate(dataset, local_snapshots, remote_dataset, remote_snapshots, replicate2_settings)
+                        # Clean snapshots remotely if one has been taken - only kept snapshots will allow aging
+                        if (dataset_settings['remote2_schema']):
+                            Cleaner.clean(remote_dataset, remote_snapshots, dataset_settings['remote2_schema'], log_command=log_command,
+                                    all_snapshots=dataset_settings['remote2_clean_all'])
+                    # Post execution command
+                    if ((result or result2) and dataset_settings['replicate_postexec'] is not None):
+                        Helper.run_command(dataset_settings['replicate_postexec'], '/', log_command=log_command)
                 else:
                     # Pull logic for remote site
                     # Replicating, if required
                     # If network replicating, check connectivity here
-                    test_unconnected = is_connected.test_unconnected(dataset_settings, local_dataset=dataset)
+                    test_unconnected = is_connected.test_unconnected(replicate_settings, local_dataset=dataset)
                     if test_unconnected:
                         log_warn("[{$0}] - Skipping as '{1}:{2}' unreachable"
                                 .format(dataset, replicate_settings['endpoint_host'], replicate_settings['endpoint_port']))
@@ -356,12 +419,11 @@ class Manager(object):
                     
                     remote_dataset = replicate_settings['target'] if push else replicate_settings['source']
                     remote_datasets = ZFS.get_datasets(replicate_settings['endpoint'], log_command=log_command)
-                    remote_snapshots = ZFS.get_snapshots(remote_dataset, replicate_settings['endpoint'], log_command=log_command,
-                            all_snapshots=dataset_settings['all_snapshots'])
                     if remote_dataset not in remote_datasets:
                         log_error("[{0}] - remote dataset '{1}' does not exist".format(dataset, remote_dataset))
                         continue
-                    remote_snapshots = remote_snapshots.get(remote_dataset, OrderedDict())
+                    remote_snapshots = ZFS.get_snapshots2(remote_dataset, replicate_settings['endpoint'], log_command=log_command,
+                            all_snapshots=dataset_settings['all_snapshots'])
                     endpoint = replicate_settings['endpoint']
                     if (take_snapshot is True and this_time not in remote_snapshots):
                         # Only execute everything here if needed
